from __future__ import annotations
from typing import List, Tuple


class Token:
    def __init__(self, fmt: str, text: str):
        self.fmt = fmt
        self.text = text
    
    def __len__(self):
        return len(self.text)

class Sentence:
    def __init__(self):
        self.data: List[Token] = []
    
    def addf(self, fmt: str, text: str):
        self.data.append(Token(fmt, text))
        return self
    
    def addt(self, text: str):
        self.data.append(Token("", text))
        return self
    
    def __getitem__(self, index: int):
        return self.data[index]
    
    def __setitem__(self, index: int, value: Token):
        self.data[index] = value
        return None

    def adds(self, fmt, sentence):
        for _, t in sentence.data:
            self.data.append(Token(fmt, t))
        return self
    
    def concat(self, sentence: Sentence):
        self.data += sentence.data
        return self
    
    def ljust(self, width: int, fillchar: str = " ", fmt = ""):
        total = self.len()
        if total < width:
            self.addf(fmt, fillchar * (width - total))
        return self
    
    def rjust(self, width: int, fillchar: str = " ", fmt = ""):
        total = self.len()
        if total < width:
            self.data = [Token(fmt, fillchar * (width - total))] + self.data
        return self
    
    def center(self, width: int, fillchar: str = " ", fmt = ""):
        total = self.len()
        if total < width:
            left = (width - total) // 2
            right = width - total - left
            self.data = [Token(fmt, fillchar * left)] + self.data + [Token(fmt, fillchar * right)]
        return self
    
    def len(self):
        total = 0
        for t in self.data:
            total += len(t.text)
        return total
    
    def __len__(self):
        return self.len()
    
    def get(self):
        return self.data

    def trim_alfa(self, limit: int):
        if limit < 0:
            return
        index = len(self.data) - 1
        size = self.len()
        while True:
            if index < 0 or size <= limit:
                break
            token = self.data[index]
            if len(token.text) == 0:
                del self.data[index]
                index -= 1
            elif token.text[-1].isalpha():
                token.text = token.text[:-1]
                size -= 1
            else:
                index -= 1

    def trim_spaces(self, limit: int):
        if limit < 0:
            return
        index = len(self.data) - 1
        size = self.len()
        while True:
            if index < 0 or size <= limit:
                break
            token = self.data[index]
            if len(token.text) == 0:
                del self.data[index]
                index -= 1
            elif token.text[-1] == " ":
                token.text = token.text[:-1]
                size -= 1
            else:
                index -= 1

    def trim_end(self, width: int):
        if width < 0:
            return
        size = self.len()
        index = len(self.data) - 1
        while True:
            if size <= width or index < 0:
                break
            token = self.data[index]
            if len(token.text) == 0:
                del self.data[index]
                index -= 1
            else:
                token.text = token.text[:-1]
                size -= 1



from .fmt import Fmt
from ..util.sentence import Sentence



class Frame:
    def __init__(self, y: int = 0, x: int = 0):
        self._x = x
        self._y = y
        self._inner_dx = 0
        self._inner_dy = 0
        self._border = "rounded"
        self._filled = False
        self._header: Sentence = Sentence()
        self._header_align = ""
        self._footer: Sentence = Sentence()
        self._footer_align = ""
        self._fill_char = " "
        self._wrap = False
        self._border_color = ""
        self._print_index = 0

    def set_border_color(self, color: str):
        self._border_color = color

    def get_dx(self):
        return self._inner_dx
    
    def get_dy(self):
        return self._inner_dy
    
    def get_x(self):
        return self._x

    def get_y(self):
        return self._y
    
    def get_header(self):
        return self._header
    
    def get_footer(self):
        return self._footer

    def set_pos(self, y: int, x: int):
        self._x = x
        self._y = y
        return self

    def set_size(self, size_y, size_x):
        self._inner_dx = size_x - 2
        self._inner_dy = size_y - 2
        return self

    def set_inner(self, inner_dy: int, inner_dx: int):
        self._inner_dy = inner_dy
        self._inner_dx = inner_dx
        return self
    
    def get_inner(self):
        return (self._inner_dy, self._inner_dx)

    def get_size(self):
        return self._inner_dy + 2, self._inner_dx + 2

    def set_end(self, y: int, x: int):
        self._inner_dx = x - self._x - 1
        self._inner_dy = y - self._y - 1
        # Fmt.write(0, 0, Sentence().addt(f"{self._inner_w}, {self._inner_h}"))
        # Fmt.getch()
        return self
    
    def set_wrap(self):
        self._wrap = True

    def set_fill_char(self, char: str):
        self._fill_char = char
        return self
    
    def get_symbol(self, value: str):
        square = { "lu": "┌", "ru": "┐", "ld": "└", "rd": "┘", "h": "─", "v": "│" }
        rounded = { "lu": "╭", "ru": "╮", "ld": "╰", "rd": "╯", "h": "─", "v": "│" }
        bold = { "lu": "┏", "ru": "┓", "ld": "┗", "rd": "┛", "h": "━", "v": "┃" }

        if self._border == "square":
            return square[value]
        elif self._border == "rounded":
            return rounded[value]
        elif self._border == "bold":
            return bold[value]
        return " "
        
    def set_border_none(self):
        self._border = "none"
        return self
        
    def set_border_bold(self):
        self._border = "bold"
        return self

    def set_border_rounded(self):
        self._border = "rounded"
        return self

    def set_border_square(self):
        self._border = "square"
        return self
 
    def set_header(self, header: Sentence, align="<"):
        self._header_align = align
        self._header = header
        return self

    def set_footer(self, footer: Sentence, align=">"):
        self._footer_align = align
        self._footer = footer
        return self

    def set_fill(self):
        self._filled = True
        return self

    def set_nofill(self):
        self._filled = False
        return self
    
    def print(self, x: int, sentence: Sentence):
        self.write(self._print_index, x, sentence)
        self._print_index += 1
        return self

    def write(self, y: int, x: int, sentence: Sentence) -> bool:
        lines, cols = Fmt.get_size()

        x_min = max(-1, self._x)
        y_min = max(-1, self._y)
        x_max = min(cols, self._x + self._inner_dx)
        y_max = min(lines, self._y + self._inner_dy)

        x_abs = x + self._x + 1
        y_abs = y + self._y + 1

        count = 0

        if y_abs <= y_min or y_abs > y_max:
            return False

        for token in sentence.get():
            fmt, text = token.fmt, token.text
            if x_abs - 1 < x_min: # Se o texto começa fora do frame
                if x_abs + len(text) > x_min: # mas ter parte dentro
                    text = text[x_min - x_abs + 1:]
                    x_abs = x_min + 1

            if x_abs <= x_max: # Se o texto começa dentro do frame
                missing = ""
                if x_abs + len(text) >= x_max:
                    cut_point = x_max - x_abs + 1
                    missing = text[cut_point:]
                    text = text[:cut_point]

                Fmt.stroke(y_abs, x_abs, fmt, text)

                if self._wrap and missing != "":
                    Fmt.stroke(y_abs + 1, x_abs, missing)
                count += 1
            x_abs += len(text)
        return False if count == 0 else True

    def draw(self):
        x = self._x
        y = self._y
        dx = self._inner_dx
        dy = self._inner_dy
        color = self._border_color
        up_left = self.get_symbol("lu")
        up_right = self.get_symbol("ru")
        down_left = self.get_symbol("ld")
        down_right = self.get_symbol("rd")
        hor = self.get_symbol("h")
        ver = self.get_symbol("v")
        header = self._header
        footer = self._footer
        if self._footer_align == "<":
            footer.ljust(dx, hor, color)
        elif self._footer_align == ">":
            footer.rjust(dx, hor, color)
        else:
            footer.center(dx, hor, color) 
    
        if self._header_align == "<":
            header.ljust(dx, hor, color)
        elif self._header_align == ">":
            header.rjust(dx, hor, color)
        else:
            header.center(dx, hor, color)

        above = Sentence().addf(color, up_left).concat(header).addf(color, up_right)
        below = Sentence().addf(color, down_left).concat(footer).addf(color, down_right)
        
        Fmt.write(y, x, above)
        Fmt.write(y + dy + 1, x, below)
        if self._filled:
            for i in range(1, dy + 1):
                Fmt.write(y + i, x, Sentence().addf(color, ver).addt(dx * self._fill_char).addf(color, ver))
        else:
            for i in range(1, dy + 1):
                Fmt.write(y + i, x, Sentence().addf(color, ver))
                Fmt.write(y + i, x + dx + 1, Sentence().addf(color, ver))
        return self



from typing import Tuple, Optional, List
import os
import urllib.request
import urllib.error
import json

from .settings import SettingsParser
from .game.game import Game
from .util.remote import RemoteCfg

class Down:
    fnprint = print

    ts_draft = (r'let _cin_ : string[] = [];' + '\n'
                r'try { _cin_ = require("fs").readFileSync(0).toString().split(/\r?\n/); } catch(e){}' + '\n'
                r'let input = () : string => _cin_.length === 0 ? "" : _cin_.shift()!;' + '\n'
                r'let write = (text: any, end:string="\n")=> process.stdout.write("" + text + end);' + '\n')

    js_draft = (r'let __lines = require("fs").readFileSync(0).toString().split("\n");' + '\n'
                r'let input = () => __lines.length === 0 ? "" : __lines.shift();' + '\n'
                r'let write = (text, end="\n") => process.stdout.write("" + text + end);') + '\n'
    
    c_draft = '#include <stdio.h>\n\nint main() {\n    return 0;\n}\n\n'
    cpp_draft = '#include <iostream>\n\nint main() {\n}\n\n'

    drafts = {'c': c_draft, 'cpp': cpp_draft, 'ts': ts_draft, 'js': js_draft}
    # def __init__(self):
    #     self.drafts = {}
    #     self.drafts['c'] = Down.c_draft
    #     self.drafts['cpp'] = Down.cpp_draft
    #     self.drafts['ts'] = Down.ts_draft
    #     self.drafts['js'] = Down.js_draft

    # @staticmethod
    # def update():
    #     if os.path.isfile(".info"):
    #         data = open(".info", "r").read().split("\n")[0]
    #         data = data.split(" ")
    #         discp = data[0]
    #         label = data[1]
    #         ext = data[2]
    #         Down.entry_unpack(".", discp, label, ext)
    #     else:
    #         print("No .info file found, skipping update...")

    @staticmethod
    def __create_file(content, path, label=""):
        with open(path, "w") as f:
            f.write(content)
        Down.fnprint("  " + path, label)

    @staticmethod
    def __unpack_json(loaded, destiny, lang: str):
        # extracting all files to folder
        for entry in loaded["upload"]:
            if entry["name"] == "vpl_evaluate.cases":
                Down.__compare_and_save(entry["contents"], os.path.join(destiny, "cases.tio"))

        # for entry in loaded["keep"]:
        #    Down.compare_and_save(entry["contents"], os.path.join(destiny, entry["name"]))

        # for entry in loaded["required"]:
        #    path = os.path.join(destiny, entry["name"])
        #    Down.compare_and_save(entry["contents"], path)

        if "draft" in loaded:
            if lang in loaded["draft"]:
                for file in loaded["draft"][lang]:
                    path = os.path.join(destiny, file["name"])
                    Down.__create_file(file["contents"], path, "(Draft)")

    @staticmethod
    def __compare_and_save(content, path):
        if not os.path.exists(path):
            with open(path, "w", encoding="utf-8") as f:
                f.write(content.encode("utf-8").decode("utf-8"))
            Down.fnprint("  " + path + " (New)")
        else:
            if open(path).read() != content:
                Down.fnprint(path + " (Updated)")
                with open(path, "w") as f:
                    f.write(content)
            else:
                Down.fnprint("  " + path + " (Unchanged)")
    
    @staticmethod
    def __down_problem_def(destiny, cache_url) -> Tuple[str, str]:
        # downloading Readme
        readme = os.path.join(destiny, "Readme.md")
        [tempfile, __content] = urllib.request.urlretrieve(cache_url + "Readme.md")

        # content = ""
        try:
            content = open(tempfile, encoding="utf-8").read()
        except FileNotFoundError:
            content = open(tempfile).read()

        Down.__compare_and_save(content, readme)
        
        # downloading mapi
        mapi = os.path.join(destiny, "mapi.json")
        urllib.request.urlretrieve(cache_url + "mapi.json", mapi)
        return readme, mapi

    @staticmethod
    def __create_problem_folder(rootdir: str, activity: str) -> str:
        # create dir
        destiny: str = os.path.join(rootdir, activity)
        if not os.path.exists(destiny):
            os.makedirs(destiny, exist_ok=True)
        else:
            Down.fnprint("  Problem folder "+ destiny + " found, merging content.")

        return destiny

    @staticmethod
    def download_problem(rootdir, course: str, activity: str, language: Optional[str], fnprint) -> bool:
        Down.fnprint = fnprint
        sp = SettingsParser()
        settings = sp.load_settings()
        rep = settings.get_repo(course)

        file = rep.get_file()
        game = Game(file)
        item = game.get_task(activity)
        if not item.link.startswith("http"):
            Down.fnprint("fail: link for activity is not a remote link")
            return False
        cfg = RemoteCfg(item.link)
        cache_url = os.path.dirname(cfg.get_raw_url()) + "/.cache/"

        destiny = Down.__create_problem_folder(rootdir, activity)
        destiny = os.path.abspath(destiny)
        try:
            [_readme_path, mapi_path] = Down.__down_problem_def(destiny, cache_url)
        except urllib.error.HTTPError:
            Down.fnprint("  fail: activity not found in course url")
            # verifi if destiny folder is empty and remove it
            if len(os.listdir(destiny)) == 0:
                os.rmdir(destiny)
            return False

        with open(mapi_path) as f:
            loaded_json = json.load(f)
        os.remove(mapi_path)

        language_def = SettingsParser().get_language()
        ask_ext = False
        if language is None:
            if language_def != "":
                language = language_def
            else:
                print("  Choose extension for draft: [c, cpp, py, ts, js, java]: ", end="")
                language = input()
                ask_ext = True
        
        Down.__unpack_json(loaded_json, destiny, language)
        Down.__download_drafts(loaded_json, destiny, language, cache_url, ask_ext)
        return True

    @staticmethod
    def __download_drafts(loaded_json, destiny: str, language, cache_url, ask_ext):
        if len(loaded_json["required"]) == 1:  # you already have the students file
            return

        if "draft" in loaded_json and language in loaded_json["draft"]:
            pass
        else:
            try:
                draft_path = os.path.join(destiny, "draft." + language)
                urllib.request.urlretrieve(cache_url + "draft." + language, draft_path)
                Down.fnprint("  " + draft_path + " (Draft) Rename before modify.")

            except urllib.error.HTTPError:  # draft not found
                filename = "draft."
                draft_path = os.path.join(destiny, filename + language)
                if not os.path.exists(draft_path):
                    with open(draft_path, "w") as f:
                        if language in Down.drafts:
                            f.write(Down.drafts[language])
                        else:
                            f.write("")
                    Down.fnprint("  " + draft_path + " (Empty)")
        
        if ask_ext:
            print("\nYou can choose default extension with command\n$ tko config -l <extension>")



from typing import List, Optional
import math
import os

from .basic import IdentifierType, Identifier, Unit, Param
from .loader import Loader
from .solver import Solver

from ..util.term_color import colour
from ..util.symbols import symbols

# generate label for cases


class LabelFactory:
    def __init__(self):
        self._label = ""
        self._index = -1

    def index(self, value: int):
        try:
            self._index = int(value)
        except ValueError:
            raise ValueError("Index on label must be a integer")
        return self

    def label(self, value: str):
        self._label = value
        return self

    def generate(self):
        label = LabelFactory.trim_spaces(self._label)
        label = LabelFactory.remove_old_index(label)
        if self._index != -1:
            index = str(self._index).zfill(2)
            if label != "":
                return index + " " + label
            else:
                return index
        return label

    @staticmethod
    def trim_spaces(text):
        parts = text.split(" ")
        parts = [word for word in parts if word != '']
        return " ".join(parts)

    @staticmethod
    def remove_old_index(label):
        split_label = label.split(" ")
        if len(split_label) > 0:
            try:
                int(split_label[0])
                return " ".join(split_label[1:])
            except ValueError:
                return label


class Wdir:
    def __init__(self):
        self.solver: Optional[Solver] = None
        self.source_list: List[str] = []
        self.pack_list: List[List[Unit]] = []
        self.unit_list: List[Unit] = []

    def set_solver(self, solver_list: List[str]):
        if len(solver_list) > 0:
            self.solver = Solver(solver_list)
        return self

    def set_sources(self, source_list: List[str]):
        self.source_list = source_list
        return self

    def set_target_list(self, target_list: List[str]):
        target_list = [t for t in target_list if t != ""]
        for target in target_list:
            if not os.path.exists(target):
                raise FileNotFoundError(colour("red", "fail: ") + target + " not found")

        solvers = [target for target in target_list if Identifier.get_type(target) == IdentifierType.SOLVER]
        sources = [target for target in target_list if Identifier.get_type(target) != IdentifierType.SOLVER]
        
        self.set_solver(solvers)
        self.set_sources(sources)
        return self

    def set_cmd(self, exec_cmd: Optional[str]):
        if exec_cmd is None:
            return self
        if self.solver is not None:
            print("fail: if using --cmd, don't pass source files to target")
        self.solver = Solver([])
        self.solver.executable = exec_cmd
        return self

    def build(self):
        loading_failures = 0
        for source in self.source_list:
            try:
                self.pack_list.append(Loader.parse_source(source))
            except FileNotFoundError as e:
                print(str(e))
                loading_failures += 1
                pass
        if loading_failures > 0 and loading_failures == len(self.source_list):
            raise FileNotFoundError("failure: none source found")
        self.unit_list = sum(self.pack_list, [])
        self.__number_and_mark_duplicated()
        self.__calculate_grade()
        self.__pad()
        return self

    def calc_grade(self) -> int:
        grade = 100
        for case in self.unit_list:
            if not case.repeated and (case.user is None or case.output != case.user):
                grade -= case.grade_reduction
        return max(0, grade)

    # put all the labels with the same length
    def __pad(self):
        if len(self.unit_list) == 0:
            return
        max_case = max([len(x.case) for x in self.unit_list])
        max_source = max([len(x.source) for x in self.unit_list])
        for unit in self.unit_list:
            unit.case_pad = max_case
            unit.source_pad = max_source

    # select a single unit to execute exclusively
    def filter(self, param: Param.Basic):
        index = param.index
        if index is not None:
            if 0 <= index < len(self.unit_list):
                self.unit_list = [self.unit_list[index]]
            else:
                raise ValueError("Index Number out of bounds: " + str(index))
        return self

    # calculate the grade reduction for the cases without grade
    # the grade is proportional to the number of unique cases
    def __calculate_grade(self):
        unique_count = len([x for x in self.unit_list if not x.repeated])
        for unit in self.unit_list:
            if unit.grade is None:
                unit.grade_reduction = math.floor(100 / unique_count)
            else:
                unit.grade_reduction = unit.grade

    # number the cases and mark the repeated
    def __number_and_mark_duplicated(self):
        new_list = []
        index = 0
        for unit in self.unit_list:
            unit.index = index
            index += 1
            search = [x for x in new_list if x.input == unit.input]
            if len(search) > 0:
                unit.repeated = search[0].index
            new_list.append(unit)
        self.unit_list = new_list

    # sort, unlabel ou rename using the param received
    def manipulate(self, param: Param.Manip):
        # filtering marked repeated
        self.unit_list = [unit for unit in self.unit_list if unit.repeated is None]
        if param.to_sort:
            self.unit_list.sort(key=lambda v: len(v.input))
        if param.unlabel:
            for unit in self.unit_list:
                unit.case = ""
        if param.to_number:
            number = 00
            for unit in self.unit_list:
                unit.case = LabelFactory().label(unit.case).index(number).generate()
                number += 1

    def unit_list_resume(self):
        return "\n".join([str(unit) for unit in self.unit_list])

    def resume(self) -> str:

        def sources() -> str:
            out = []
            if len(self.pack_list) == 0:
                out.append(symbols.failure)
            for i in range(len(self.pack_list)):
                nome: str = self.source_list[i].split(os.sep)[-1]
                out.append(nome + "(" + str(len(self.pack_list[i])).zfill(2) + ")")
            return colour("green", "base:") + "[" + ", ".join(out) + "]"

        def solvers() -> str:
            path_list = [] if self.solver is None else self.solver.path_list

            if self.solver is not None and len(path_list) == 0:  # free_cmd
                out = "free cmd"
            else:
                out = ", ".join([os.path.basename(path) for path in path_list])
            return colour("green", "prog:") + "[" + out + "]"

        # folder = os.getcwd().split(os.sep)[-1]
        # tests_count = (colour("tests:", Color.GREEN) +
        #               str(len([x for x in self.unit_list if x.repeated is None])).zfill(2))

        return symbols.opening + sources() + " " + solvers()



class Style:
    focus = "Y"
    italic = "/"
    underline = "_"

    opt_quest = italic + "m"
    opt_task = italic + "c"
    text = ""
    cluster_key = "b"
    cluster_title = ""
    quest_key = italic + "b"
    tasks = "y"
    lcmd = 'r'
    cmd = ""
    code_key = ""
    skills = "c"
    play = "g"
    new = "g"

    nothing = "m"
    started = "r"
    required = "y"
    complete = "g"

    shell = "r" # extern shell cmds
    htext = ""
    check = "g"
    uncheck = "y"
    param = "c"

    bar_main = "M"
    bar_ext = "Y"
    bar_xp = "C"
    flags_true = "G"
    flags_false = "Y"
    bar_skills = "B"
    bar_cmds = "Y"
    bar_bg = "M"


tko_guide = """
       ╔════ TKO GUIA COMPACTO ════╗
╔══════╩═════ BAIXAR PROBLEMA ═════╩═══════╗
║        tko down <curso> <label>          ║
║ exemplo poo  : tko down poo carro        ║
║ exemplo fup  : tko down fup opala        ║
╟─────────── EXECUTAR SEM TESTAR ──────────╢
║          tko run <cod, cod...>           ║
║exemplo ts  : tko run solver.ts           ║
║exemplo cpp : tko run main.cpp lib.cpp    ║
╟──────────── RODAR OS TESTES ─────────────╢
║   tko run cases.tio <cod, ...> [-i ind]  ║
║ exemplo: tko run cases.tio main.ts       ║
║só ind 6: tko run cases.tio main.c -i 6   ║
╟── DEFINIR EXTENSÃO PADRÃO DOS RASCUNHOS ─╢
║           tko config -l <ext>            ║
║     exemplo c : tko config -l c          ║
║  exemplo java : tko config -l java       ║
╟─────────── MUDAR VISUALIZAÇÃO ───────────╢
║             tko config <--opcao>         ║
║DiffMode: tko config [--side  | --updown ]║
║Cores   : tko config [--mono  | --color  ]║
║Encoding: tko config [--ascii | --unicode]║
╚══════════════════════════════════════════╝
"""

bash_guide = """
       ╔═══ BASH  GUIA COMPACTO ════╗
╔══════╩════ MOSTRAR E NAVEGAR ═════╩══════╗
║Mostrar arquivos  : ls                    ║
║Mostrar ocultos   : ls -la                ║
║Mudar de pasta    : cd _nome_da_pasta     ║
║Subir um nível    : cd ..                 ║
╟─────────────── CRIAR ────────────────────╢
║Criar um arquivo  : touch _nome_do_arquivo║
║Criar uma pasta   : mkdir _nome_da_pasta  ║
╟─────────────── REMOVER ──────────────────╢
║Apagar um arquivo : rm _nome_do_arquivo   ║
║Apagar uma pasta  : rm -r _nome_da_pasta  ║
║Renomear ou mover : mv _antigo _novo      ║
╟─────────────── CONTROLAR ────────────────╢
║Últimos comandos  : SETA PRA CIMA         ║
║Limpar console    : Control L             ║
║Cancelar execução : Control C             ║
║Finalizar entrada : Control D             ║
╚══════════════════════════════════════════╝
"""


from typing import List, Tuple
from .frame import Frame
from ..util.sentence import Sentence
from .fmt import Fmt

import curses
import time

class Input:
    def __init__(self):
        self.frame = Frame(0, 0)
        self.content: List[Sentence] = []
        # self.ending = 0
        # self.total_time = 0
        self.type = ""
        self.options = []
        self.options_index = 0
        self.fn_answer = None
        self.enable = True
        self.extra_exit: List[int] = []
        self.exit_fn = None
        self.exit_key = None

    def set_header(self, text: Sentence, align: str = '<'):
        self.frame.set_header(text, align)
        return self
    
    def set_exit_key(self, key: str):
        self.exit_key = ord(key)
        return self

    def set_exit_fn(self, fn):
        self.exit_fn = fn
        return self

    def add_extra_exit(self, key: str):
        self.extra_exit.append(ord(key[0]))
        return self

    def __setup_frame(self):
        lines, cols = Fmt.get_size()
        max_dx = max([x.len() for x in self.content])
        dx = max_dx + 2
        dy = len(self.content) + 2
        x = (cols - dx) // 2
        y = (lines - dy) // 2
        self.frame.set_pos(y, x).set_inner(dy, dx)
        self.frame.set_fill()
        # if self.type == "timer":
        #     delta = dx - 4
        #     percent = (self.ending - time.time()) / self.total_time
        #     footer = (delta - int(delta * percent)) * "─" + int(delta * percent) * "━"
        #     self.frame.set_footer(Sentence().addt(footer), "^")
        
        if self.type == "answer":
            footer = Sentence().addt(" ")
            for i, option in enumerate(self.options):
                fmt = "kG" if i == self.options_index else ""
                footer.addf(fmt, option).addt(" ")
            self.frame.set_footer(footer, "^")

    def addt(self, text: str):
        self.content.append(Sentence().addt(text))
        return self

    def adds(self, sentence: Sentence):
        self.content.append(sentence)
        return self
    
    def set_content(self, content: List[str]):
        self.content = [Sentence().addt(x) for x in content]
        return self

    # def timer(self, delta: int):
    #     self.type = "timer"
    #     self.frame.set_border_rounded()
    #     self.frame.set_header(Sentence().addf("/", " Aviso "))
    #     self.enable = True
    #     self.total_time = delta
    #     self.ending = time.time() + delta

    #     return self

    def warning(self):
        self.type = "warning"
        if self.frame.get_header().len() == 0:
            self.frame.set_header(Sentence().addf("/", " Aviso "))
        self.frame.set_footer(Sentence().addf("/", " Pressione enter "))
        self.frame.set_border_color("y")
        return self
    
    def error(self):
        self.type = "warning"
        if self.frame.get_header().len() == 0:
            self.frame.set_header(Sentence().addf("/", " Erro "))
        self.frame.set_footer(Sentence().addf("/", " Pressione enter "))
        self.frame.set_border_color("r")
        return self
    
    def set_options(self, options: List[str]):
        self.options = options
        return self
    
    def get_answer(self, fn_answer):
        self.type = "answer"
        self.frame.set_border_rounded()
        self.frame.set_header(Sentence().addf("/", " Pergunta "))
        # self.frame.set_footer(Sentence().addf("/", " Selecione e aperte enter "))
        self.fn_answer = fn_answer
        return self

    def draw(self):
        self.__setup_frame()
        self.frame.draw()
        for i, line in enumerate(self.content):
            self.frame.write(i + 1, 1, line)
        return self

    def get_input(self) -> int:
        self.draw()
        key = Fmt.getch()
        if self.type == "warning" or self.type == "answer":
            if key == ord('\n') or key == 27 or key in self.extra_exit:
                self.enable = False
                if self.exit_fn is not None:
                    self.exit_fn()
                if self.exit_key is not None:
                    return self.exit_key
        if self.type == "answer":
            if key == curses.KEY_LEFT:
                self.options_index = (self.options_index - 1) % len(self.options)
            elif key == curses.KEY_RIGHT:
                self.options_index = (self.options_index + 1) % len(self.options)
            elif key == ord('\n'):
                self.enable = False
                if self.fn_answer is not None:
                    self.fn_answer(self.options[self.options_index])
                if self.exit_fn is not None:
                    self.exit_fn()
                if self.exit_key is not None:
                    return self.exit_key
                return -1
        return -1
        


from ..game.game import Game
from ..game.cluster import Cluster
from ..game.quest import Quest
from ..game.task import Task
from ..game.xp import XP
from ..game.graph import Graph

from typing import List, Dict, Tuple, Optional, Any
from ..settings import RepoSettings, LocalSettings
from ..down import Down
from ..util.symbols import symbols
from ..util.sentence import Sentence
from .style import Style
from .fmt import Fmt
from .frame import Frame
from .input import Input
import time
import webbrowser

import os
import curses

class Entry:
    def __init__(self, obj: Any, sentence: Sentence):
        self.obj = obj
        self.sentence = sentence

    def get(self):
        return self.sentence

class Flags:
    count = "count"
    percent = "percent"
    goals = "goals"
    down = "down"
    admin = "admin"
    flags_bar = "flags_bar"
    skills_bar = "skills_bar"
    help_bar = "help_bar"
    top_bar = "top_bar"

class Play:
    cluster_prefix = "'"

    def __init__(self, local: LocalSettings, game: Game, rep: RepoSettings, repo_alias: str, fnsave):
        self.fnsave = fnsave
        self.local = local
        self.repo_alias = repo_alias
        self.rep = rep

        self.input_layer: List[Input] = []

        # self.mark_opt = "opt" in self.rep.view
        self.mark_opt = True
        order = [entry for entry in self.rep.view if entry.startswith("order:")]
        if len(order) > 0:
            self.flags = [v for v in order[0][6:].split(",") if v != '']
        else:
            self.flags = []

        self.game: Game = game
        self.expanded: List[str] = [x for x in self.rep.expanded]
        self.new_items: List[str] = [x for x in self.rep.new_items]
        self.avaliable_quests: List[Quest] = [] # avaliable quests
        self.avaliable_clusters: List[Cluster] = [] # avaliable clusters

        self.index_selected = 0
        self.index_begin = 0
  
        
        self.items: List[Entry] = []

        self.first_loop = True
        self.graph_ext = ""

        self.load_rep()

        self.help_base = ["Quit[q]", "Help[h]", ]
        self.help_extra = ["Move[wasd]", "Mark[enter]", "Grade[0-9]", "OpenLink[o]", "GetTask[g]", "Lang[l]", "Expand[>]", "Collapse[<]",  "MassMark[M]"]

    def save_to_json(self):
        self.rep.expanded = [x for x in self.expanded]
        self.rep.new_items = [x for x in self.new_items]

        self.rep.tasks = {}
        for t in self.game.tasks.values():
            if t.grade != 0:
                self.rep.tasks[t.key] = str(t.grade)
        self.rep.view = []
        self.rep.view.append("order:" + ",".join(self.flags))
            
        self.fnsave()

    def test_color(self, scr):
        fg = "rgbymcwk"
        bg = " RGBYMCWK"
        lines, cols = scr.getmaxyx()
        scr.erase()
        for l in range(8):
            for c in range(9):
                cor = fg[l] + bg[c]
                Fmt.write(3 * l, 5 * c, Sentence().addf(cor, f" {cor} "))
        
        scr.refresh()
        scr.getch()

    def load_rep(self):
        for key, grade in self.rep.tasks.items():
            if key in self.game.tasks:
                value = "0"
                if grade == "x":
                    value = "10"
                elif grade == "":
                    value = "0"
                else:
                    value = grade
                self.game.tasks[key].set_grade(int(value))

    def update_avaliable_quests(self):
        old_quests = [q for q in self.avaliable_quests]
        old_clusters = [c for c in self.avaliable_clusters]
        if self.flags_on(Flags.admin):
            self.avaliable_quests = list(self.game.quests.values())
            self.avaliable_clusters = self.game.clusters
        else:
            self.avaliable_quests = self.game.get_reachable_quests()
            self.avaliable_clusters = []
            for c in self.game.clusters:
                if any([q in self.avaliable_quests for q in c.quests]):
                    self.avaliable_clusters.append(c)


        removed_clusters = [c for c in old_clusters if c not in self.avaliable_clusters]
        for c in removed_clusters:
            if c.key in self.expanded:
                self.expanded.remove(c.key)
        removed_quests = [q for q in old_quests if q not in self.avaliable_quests]
        for q in removed_quests:
            if q.key in self.expanded:
                self.expanded.remove(q.key)

        added_clusters = [c for c in self.avaliable_clusters if c not in old_clusters]
        added_quests = [q for q in self.avaliable_quests if q not in old_quests]
        
        for c in added_clusters:
            self.new_items.append(c.key)
        for q in added_quests:
            self.new_items.append(q.key)

    def add_focus(self, color, in_focus):
        if not in_focus:
            return color
        if color == "":
            return "k" + Style.focus
        return Style.focus + color

    def str_task(self, in_focus: bool, t: Task, ligc: str, ligq: str, min_value = 1) -> Sentence:
        output = Sentence()
        output.addt(" " + ligc + " " + ligq)\
              .concat(t.get_grade_symbol(min_value))\
              .addt(" ")

        value = Style.opt_task if self.mark_opt and t.opt else ""
        output.addf(self.add_focus(value, in_focus), t.title)

        if self.flags_on(Flags.down):
            path = os.path.join(self.local.rootdir, self.repo_alias, t.key)
            if os.path.isdir(path):
                output.addt(" ").addf("y", f"[{path}]")

        if self.flags_on(Flags.skills_bar):
            xp = ""
            for s, v in t.skills.items():
                xp += f" +{s}:{v}"
            output.addf(Style.skills, xp)
        return output
    
    def str_quest(self, in_focus: bool, q: Quest, lig: str) -> Sentence:
        con = "━─" if q.key not in self.expanded else "─┯"
        output: Sentence = Sentence().addt(" " + lig + con + " ")

        value = Style.opt_quest if self.mark_opt and q.opt else ""
        output.addf(self.add_focus(value, in_focus), q.title)

        for item in self.flags:
            if item == Flags.count:
                output.addt(" ").concat(q.get_resume_by_tasks())
            elif item == Flags.percent:
                output.addt(" ").concat(q.get_resume_by_percent())
            elif item == Flags.goals:
                output.addt(" ").concat(q.get_requirement())
        if self.flags_on(Flags.skills_bar):
            xp = ""
            for s,v in q.skills.items():
                xp += f" +{s}:{v}"
            output.addf(Style.skills, " " + xp)
                
        if q.key in self.new_items:
            output.addf(Style.new, " [new]")

        return output
        
    def str_cluster(self, in_focus: bool, cluster: Cluster, quests: List[Quest]) -> Sentence:
        output: Sentence = Sentence()
        opening = "━─"        
        if cluster.key in self.expanded:
            opening = "─┯"        
        output.addt(opening + " ")
        value = Style.cluster_title
        output.addf(self.add_focus(value, in_focus), cluster.title.strip())
        output.addt(" ").concat(cluster.get_resume_by_percent())
        if cluster.key in self.new_items:
            output.addf(Style.new, " [new]")

        return output
    
    def get_avaliable_quests_from_cluster(self, cluster: Cluster) -> List[Quest]:
        return [q for q in cluster.quests if q in self.avaliable_quests]

    def reload_options(self):
        index = 0
        self.items = []
        for cluster in self.avaliable_clusters:
            quests = self.get_avaliable_quests_from_cluster(cluster)
            sentence = self.str_cluster(self.index_selected == index, cluster, quests)
            self.items.append(Entry(cluster, sentence))
            index += 1

            if not cluster.key in self.expanded: # va para proximo cluster
                continue

            for q in quests:
                lig = "├" if q != quests[-1] else "╰"
                sentence = self.str_quest(self.index_selected == index, q, lig)
                self.items.append(Entry(q, sentence))
                index += 1
                if q.key in self.expanded:
                    for t in q.get_tasks():
                        ligc = "│" if q != quests[-1] else " "
                        ligq = "├ " if t != q.get_tasks()[-1] else "╰ "
                        sentence = self.str_task(self.index_selected == index, t, ligc, ligq, q.tmin)
                        self.items.append(Entry(t, sentence))
                        index += 1

        if self.index_selected >= len(self.items):
            self.index_selected = len(self.items) - 1

    def down_task(self, rootdir, obj: Any, ext: str):
        if isinstance(obj, Task) and obj.key in obj.title:
            task: Task = obj
            down_frame = Input().warning().set_header(Sentence().addt(" Baixando tarefa "))
            down_frame.addt(f"tko down {self.repo_alias} {task.key} -l {ext}")
            self.add_input(down_frame)

            def fnprint(text):
                down_frame.addt(text)
                down_frame.draw()
                Fmt.refresh()

            Down.download_problem(rootdir, self.repo_alias, task.key, ext, fnprint)
        else:
            self.add_input(Input().addt("Essa não é uma tarefa de código").error())


    def set_rootdir(self):
        def chama(value):
            if value == "yes":
                self.local.rootdir = os.getcwd()
                self.fnsave()
                self.add_input(Input()\
                    .addt("Diretório raiz definido como ")\
                    .addt("  " +  self.local.rootdir)\
                    .addt("Você pode alterar o diretório raiz navegando para o")\
                    .addt("diretório desejado e executando o comando")\
                    .addt("  tko config --root")\
                    .addt("")\
                    .set_exit_fn(self.set_language)
                    .warning())
            else:
                self.add_input(Input()\
                    .addt("Navague para o diretório desejado e execute o comando")\
                    .addt("  tko config --root")\
                    .addt("ou rode o tko play novamente na pasta desejada").warning())

        
        self.add_input(Input().addt("Diretório raiz para o tko ainda não foi definido")\
            .addt("Você deseja utilizar o diretório atual")\
            .addt(os.getcwd())\
            .addt("como raiz para o repositório de " + self.repo_alias + "?")\
            .set_options(["yes", "no"])\
            .get_answer(chama))
    
    def set_language(self):

        def back(value):
            self.rep.lang = value
            self.fnsave()
            self.add_input(Input()\
            .addt("      Linguagem alterada para " + value)\
            .addt("Você pode mudar a linguagem de programação")\
            .adds(Sentence().addt("            apertando ").addf("G", "e"))\
            .warning())
            

        self.add_input(Input()\
            .addt("   Escolha a extensão")\
            .addt(" default para os rascunhos") \
            .addt("")\
            .addt(" Selecione e tecle enter")\
            .set_options(["c", "cpp", "py", "ts", "js", "java", "hs"])\
            .get_answer(back))

    def process_down(self):

        if self.local.rootdir == "":
            self.set_rootdir()
            return

        if self.rep.lang == "":
            self.set_language()
            return
    
        rootdir = os.path.relpath(os.path.join(self.local.rootdir, self.repo_alias))
        obj = self.items[self.index_selected].obj
        self.down_task(rootdir, obj, self.rep.lang)
        


    def process_collapse(self):
        quest_keys = [q.key for q in self.avaliable_quests]
        if any([q in self.expanded for q in quest_keys]):
            self.expanded = [key for key in self.expanded if key not in quest_keys]
        else:
            self.expanded = []

    def process_expand(self):
        # if any cluster outside expanded
        expand_clusters = False
        for c in self.avaliable_clusters:
            if c.key not in self.expanded:
                expand_clusters = True
        if expand_clusters:
            for c in self.avaliable_clusters:
                if c.key not in self.expanded:
                    self.expanded.append(c.key)
        else:
            for q in self.avaliable_quests:
                if q.key not in self.expanded:
                    self.expanded.append(q.key)
    
    def flags_toggle(self, token):
        if token in self.flags:
            self.flags.remove(token)
        else:
            self.flags.append(token)



    @staticmethod
    def checkbox(value) -> Sentence:
        return Sentence().addf(Style.check, symbols.opcheck) if value else Sentence().addf(Style.uncheck, symbols.opuncheck)

    def build_bar_links(self) -> str:
        obj = self.items[self.index_selected].obj
        if isinstance(obj, Task):
            link = obj.link
            if link:
                return link
        return ""

    def flags_on(self, value):
        return value in self.flags

    def flags_color(self, flag: str) -> str:
        if self.flags_on(flag):
            return Style.flags_true
        return Style.flags_false

    def build_xp_bar(self, text:str, percent: float, length: int) -> Sentence:
        prefix = (length - len(text)) // 2
        sufix = length - len(text) - prefix
        text = " " * prefix + text + " " * sufix
        xp_bar = Sentence()
        total = length
        full_line = text
        done_len = int(percent * total)
        xp_bar.addf("/kC", full_line[:done_len]).addf("/kY", full_line[done_len:])
        return xp_bar

    def show_main_bar(self, frame: Frame):
        frame.set_header(Sentence().addt("{").addf("/", f"Tarefas lang:{self.rep.lang}").addt("}"))
        frame.set_footer(Sentence().addt("{" + self.build_bar_links() + "}"))
        frame.draw()

        dy, dx = frame.get_inner()
        y = 0
        if len(self.items) < dy:
            self.index_begin = 0
        else:
            if self.index_selected < self.index_begin: # subiu na tela
                self.index_begin = self.index_selected
            elif self.index_selected >= dy + self.index_begin: # desceu na tela
                self.index_begin = self.index_selected - dy + 1
        init = self.index_begin
    
        for i in range(init, len(self.items)):
            sentence = self.items[i].sentence
            frame.write(y, 0, sentence)
            y += 1

    def show_skills_bar(self, frame_xp):
            dy, dx = frame_xp.get_inner()
            xp = XP(self.game)
            total_perc = int(100 * (xp.get_xp_total_obtained() / xp.get_xp_total_available()))
            if self.flags_on(Flags.percent):
                text =  f" Total:{total_perc}%"
            else:
                text =  f" Total:{xp.get_xp_total_obtained()}"
    

            total_bar = self.build_xp_bar(text, total_perc / 100, dx - 2)
            frame_xp.set_header(Sentence().addt("{").addf("/", "Skills").addt("}"), "^")
            frame_xp.set_footer(Sentence().concat(total_bar), "^")
            frame_xp.draw()

            total, obt = self.game.get_skills_resume()
            index = 0
            for skill, value in total.items():
                if self.flags_on(Flags.percent):
                    text = f"{skill}:{int(100 * obt[skill] / value)}%"
                else:
                    text = f"{skill}:{obt[skill]}/{value}"
    
                perc = obt[skill]/value
                skill_bar = self.build_xp_bar(text, perc, dx - 2 )
                frame_xp.write(index, 1, skill_bar)
                index += 2

    def show_flags_bar(self, frame: Frame):
        frame.set_header(Sentence().addt("{").addf("/", "Flags").addt("}"), "^")
        frame.draw()

        def add_flag(_flag, _text, _key):
            color = self.flags_color(_flag)
            if self.flags_on(_flag):
                s = Sentence().addf(color, _text).addf(color, (8 - len(_text)) * " ").addf("B", _key)
            else:
                s = Sentence().addf("B", _key).addf(color, (8 - len(_text)) * " ").addf(color, _text)
            frame.print(1, s)


        add_flag(Flags.count,   "Count",   "[C]")
        add_flag(Flags.percent, "Percent", "[P]")
        add_flag(Flags.goals,   "Goals",   "[G]")
        add_flag(Flags.down,    "Down",    "[D]")
        add_flag(Flags.admin,   "Admin",   "[A]")



    def build_list_sentence(self, items: List[str]) -> Sentence:
        help = Sentence()
        try:
            for x in items:
                label, key = x.split("[")
                key = "[" + key
                help.addf("/", label).addt(key).addt(" ")
        except ValueError:
            raise ValueError("Desempacotando mensagens")        
        help.data.pop()

        return help

    def show_help(self):
        help: Input = Input().add_extra_exit("h").warning()
        self.add_input(help)
        help.set_header(Sentence().addf("/", " Help "))
        help.addt("Barras alternáveis")
        help.adds(Sentence().addt("  ").addf("", "F").addt(" - Mostrar ou esconder a barra de flags"))
        help.adds(Sentence().addt("  ").addf("", "H").addt(" - Mostrar ou esconder a barra de ajuda"))
        help.adds(Sentence().addt("  ").addf("", "S").addt(" - Mostrar ou esconder a barra de skills"))
        help.addt("")
        help.addt("Comandos")
        help.addt("  q - Fechar o aplicativo")
        help.addt("  h - Mostrar essa tela de ajuda")
        help.addt("  l - Mudar a linguagem de download dos rascunhos")
        help.addt("  > - Expandir tudo")
        help.addt("  < - Contrair tudo")
        help.addt("")
        help.addt("Flags")
        help.addt("  C - Contagem dos elementos finalizados")
        help.addt("  P - Porcentagem dos elementos finalizados")
        help.addt("  G - Mostrar as metas q[porcentagem] ou t[valor minimo]")
        help.addt("  D - Mostrar o caminho das tarefas baixadas")
        help.addt("  A - Modo administrador - desbloqueia todas as tarefas")
        help.addt("")
        help.addt("Controles")
        help.addt("  setas - Para navegar entre os elementos")
        help.addt("  enter - Marcar ou desmarcar")
        help.addt("  0 a 9 - Definir a nota parcial")
        help.addt("      d - Baixar a tarefa")

    def desable_on_resize(self):
        lines, cols = Fmt.get_size()
        if cols < 50 and self.flags_on("skills_bar") and self.flags_on("flags_bar"):
            self.flags_toggle("skills_bar")
        elif cols < 30 and self.flags_on("flags_bar"):
            self.flags_toggle("flags_bar")
        elif cols < 35 and self.flags_on("skills_bar"):
            self.flags_toggle("skills_bar")

    def draw_warnings(self):
        if len(self.input_layer) > 0:
            if not self.input_layer[0].enable:
                self.input_layer = self.input_layer[1:]

        if len(self.input_layer) > 0 and self.input_layer[0].enable:
            self.input_layer[0].draw()

    def show_bottom_bar(self, frame: Frame):
        help = self.build_list_sentence(self.help_extra)
        for s in help:
            if s.text.startswith("["):
                s.fmt = "B"
        dx = frame.get_dx()
        # help.trim_spaces(dx)
        help.trim_alfa(dx)
        help.trim_end(dx)
        frame.set_header(help, "^")
        # frame.set_border_square()
        frame.set_border_none()
        frame.draw()

    def show_top_bar(self, frame: Frame) -> None:
        help = self.build_list_sentence(self.help_base)
        dx = frame.get_dx()
        help.trim_alfa(dx)
        help.trim_end(dx)
        for s in help:
            if s.text.startswith("["):
                s.fmt = "B"
        frame.set_footer(help, "^")
        frame.draw()


        content = Sentence().addt(" ")
        content.addf("C", f"({self.repo_alias.upper()})").addt(" ")

        def add_flag(flag, text, key):
            color = self.flags_color(flag)
            if self.flags_on(flag):
                content.addf(color + "/", text).addf(color, key).addt(" ")
            else:
                content.addf(color, key).addf(color + "/", text).addt(" ")
        
        add_flag(Flags.flags_bar, "FlagsBar", "[F]")
        add_flag(Flags.help_bar, "HelpBar", "[H]")
        add_flag(Flags.skills_bar, "SkillsBar", "[S]")
        
        for s in content:
            if s.text.startswith("["):
                s.fmt = "B"

        xp = XP(self.game)

        if xp.get_xp_total_obtained() == xp.get_xp_total_available():
            text = "Você atingiu o máximo de xp!"
            percent = 100.0
        else:
            if self.flags_on(Flags.percent):
                text = f"L:{xp.get_level()} XP:{int(100 * xp.get_xp_level_current() / xp.get_xp_level_needed())}%"
            else:
                text = f"L:{xp.get_level()} XP:{xp.get_xp_level_current()}/{xp.get_xp_level_needed()}"
            percent = float(xp.get_xp_level_current()) / float(xp.get_xp_level_needed())
        size = max(15, dx - content.len() - 1)
        xp_bar = self.build_xp_bar(text, percent, size).addt(" ")

        limit = dx - xp_bar.len()
        content.trim_spaces(limit)
        content.trim_alfa(limit)
        content.trim_end(limit)

        frame.write(0, 0, content.concat(xp_bar))
        return content

    def show_items(self):
        self.desable_on_resize()
        self.reload_options()        
        Fmt.erase()

        lines, cols = Fmt.get_size()
        main_sx = cols  # tamanho em x livre
        main_sy = lines #size em y avaliable
        
        top_y = -1
        top_dy = 3
        frame_top = Frame(top_y, 0).set_size(3, main_sx)
        self.show_top_bar(frame_top)
        
        bottom_sy = 0
        if self.flags_on(Flags.help_bar):
            bottom_sy = 1
            frame_bottom = Frame(lines - 1, 0).set_size(3, main_sx)
            self.show_bottom_bar(frame_bottom)

        mid_y = top_y + top_dy
        mid_sy = main_sy - (top_y + top_dy + bottom_sy)

        skills_sx= 0
        if self.flags_on("skills_bar"):
            skills_sx = max(20, main_sx // 3)
            frame_skills = Frame(mid_y, cols - skills_sx).set_size(mid_sy, skills_sx)
            self.show_skills_bar(frame_skills)

        flags_sx = 0
        if self.flags_on("flags_bar"):
            flags_sx = 15
            frame_flags = Frame(mid_y, 0).set_size(mid_sy, flags_sx)
            self.show_flags_bar(frame_flags)

        task_sx = main_sx - flags_sx - skills_sx
        frame_main = Frame(mid_y, flags_sx).set_size(mid_sy, task_sx)
        self.show_main_bar(frame_main)

    def generate_graph(self, scr):
        if not self.first_loop:
            return
        if self.graph_ext == "":
            return
        reachable: List[str] = [q.key for q in self.avaliable_quests]
        counts = {}
        for q in self.game.quests.values():
            done = len([t for t in q.get_tasks() if t.is_complete()])
            init = len([t for t in q.get_tasks() if t.in_progress()])
            todo = len([t for t in q.get_tasks() if t.not_started()])
            counts[q.key] = f"{done} / {done + init + todo}\n{q.get_percent()}%"

        Graph(self.game).set_opt(self.mark_opt).set_reachable(reachable).set_counts(counts).set_graph_ext(self.graph_ext).generate()
        lines, _cols = scr.getmaxyx()
        if self.first_loop:
            text = Sentence().addt(f"Grafo gerado em graph{self.graph_ext}")
            Fmt.write(lines - 1, 0, text)
            
    def update_new(self):
        self.new_items = [item for item in self.new_items if item not in self.expanded]

    def mass_mark(self):
        obj = self.items[self.index_selected].obj
        if isinstance(obj, Cluster):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)
                return
            full_open = True
            for q in self.get_avaliable_quests_from_cluster(obj):
                if q.key not in self.expanded:
                    self.expanded.append(q.key)
                    full_open = False
            if not full_open:
                return
            
            value = None
            for q in obj.quests:
                for t in q.get_tasks():
                    if value is not None:
                        t.set_grade(value)
                    else:
                        value = 10 if t.grade < 10 else 0
                        t.set_grade(value)
        elif isinstance(obj, Quest):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)
            else:
                value = None
                for t in obj.get_tasks():
                    if value is not None:
                        t.set_grade(value)
                    else:
                        value = 10 if t.grade < 10 else 0
                        t.set_grade(value)
        else:
            obj.set_grade(10 if obj.grade < 10 else 0)

    def set_grade(self, grade):
        obj = self.items[self.index_selected].obj
        if isinstance(obj, Task):
            obj.set_grade(grade - ord("0"))

    def arrow_right(self):
        obj = self.items[self.index_selected].obj
        if isinstance(obj, Cluster):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)
            else:
                self.index_selected += 1
        elif isinstance(obj, Quest):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)
            else:
                while True:
                    self.index_selected += 1
                    obj = self.items[self.index_selected].obj
                    if isinstance(obj, Cluster) or isinstance(obj, Quest):
                        break
                    if self.index_selected == len(self.items) - 1:
                        break
        elif isinstance(obj, Task):
            while True:
                obj = self.items[self.index_selected].obj
                if isinstance(obj, Quest) or isinstance(obj, Cluster):
                    break
                if self.index_selected == len(self.items) - 1:
                    break
                self.index_selected += 1

    def arrow_left(self):
        obj = self.items[self.index_selected].obj
        if isinstance(obj, Quest):
            if obj.key in self.expanded:
                self.expanded.remove(obj.key)
            else:
                while True:
                    if self.index_selected == 0:
                        break
                    self.index_selected -= 1
                    obj = self.items[self.index_selected].obj
                    if isinstance(obj, Cluster) or isinstance(obj, Quest) and obj.key in self.expanded:
                        break
                    if self.index_selected == 0:
                        break
        elif isinstance(obj, Cluster):
            if obj.key in self.expanded:
                self.expanded.remove(obj.key)
                for q in obj.quests:
                    try:
                        self.expanded.remove(q.key)
                    except ValueError:
                        pass
            else:
                while True:
                    if self.index_selected == 0:
                        break
                    self.index_selected -= 1
                    obj = self.items[self.index_selected].obj
                    if isinstance(obj, Cluster) or isinstance(obj, Quest):
                        break
                    if self.index_selected == 0:
                        break
        elif isinstance(obj, Task):
            while True:
                obj = self.items[self.index_selected].obj
                if isinstance(obj, Quest):
                    break
                self.index_selected -= 1
            
    def expand(self):
        obj = self.items[self.index_selected].obj
        if isinstance(obj, Quest) or isinstance(obj, Cluster):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)

    def open_link(self):
        obj = self.items[self.index_selected].obj
        if isinstance(obj, Task):
            task: Task = obj
            if task.link.startswith("http"):
                webbrowser.open_new_tab(task.link)
            self.add_input(Input().set_header(Sentence().addf("/", " Abrindo link ")).addt(task.link).warning())
        else:
            self.add_input(Input().addt("O comando de abrir link só").addt("   funciona em tarefas").error())

    def toggle(self):
        obj = self.items[self.index_selected].obj
        if isinstance(obj, Task):
            if obj.grade < 10:
                obj.set_grade(10)
            else:
                obj.set_grade(0)
        elif isinstance(obj, Quest) or isinstance(obj, Cluster):
            if obj.key in self.expanded:
                self.expanded.remove(obj.key)
            else:
                self.expanded.append(obj.key)

    def add_input(self, value: Input):
        self.input_layer.append(value)

    def check_disable_flags(self):
        lines, cols = Fmt.get_size()
        if cols < 50:
            if self.flags_on("flags_bar"):
                self.flags_toggle("flags_bar")

    def check_disable_skills(self):
        lines, cols = Fmt.get_size()
        if cols < 50:
            if self.flags_on(Flags.skills_bar):
                self.flags_toggle("skills_bar")
        
    def send_ligado(self, msg: str, flag: bool, key: str) -> Sentence:
        if self.flags_on(flag):
            opt = Sentence().addt(msg).addf("G", "ligado")
        else:
            opt = Sentence().addt(msg).addf("R", "desligado")
        
        self.add_input(Input().adds(opt).warning().add_extra_exit(key))


    def main(self, scr):
        # scr.nodelay(True)
        curses.curs_set(0)  # Esconde o cursor
        Fmt.init_colors()  # Inicializa as cores
        Fmt.set_scr(scr)  # Define o scr como global

        # Exemplo de uso da função escrever
        while True:
            self.update_avaliable_quests()
            self.update_new()
            self.show_items()
            self.draw_warnings()
            self.generate_graph(scr)

            if len(self.input_layer) > 0 and self.input_layer[0].enable:
                value = self.input_layer[0].get_input()
            else:
                value = scr.getch()

            if value == ord("Q"):
                break
            elif value == ord("q") or value == 27:
                self.add_input(Input().addt("Saindo do aplicativo").set_exit_key("Q").add_extra_exit("q").warning())
            elif value == curses.KEY_UP or value == ord("w"):
                self.index_selected = max(0, self.index_selected - 1)
            elif value == curses.KEY_DOWN or value == ord("s"):
                self.index_selected = min(len(self.items) - 1, self.index_selected + 1)
            elif value == curses.KEY_LEFT or value == ord("a"):
                self.arrow_left()
            elif value == curses.KEY_RIGHT or value == ord("d"):
                self.arrow_right()
            elif value == ord("H"):
                self.flags_toggle("help_bar")
            elif value == ord("S"):
                self.flags_toggle(Flags.skills_bar)
                self.check_disable_flags()
            elif value == ord("F"):
                self.flags_toggle(Flags.flags_bar)
                self.check_disable_skills()
            elif value == ord("h"):
                self.show_help()
            elif value == ord("C"):
                self.flags_toggle(Flags.count)
                self.send_ligado("Contagem de tarefas feitas ", Flags.count, "C")
            elif value == ord("P"):
                self.flags_toggle(Flags.percent)
                self.send_ligado("Mostrando percentuais ", Flags.percent, "P")
            elif value == ord("G"):
                self.flags_toggle(Flags.goals)
                self.send_ligado("Mostrando metas para conclusão de Missão ", Flags.goals, "G")
            elif value == ord("D"):
                self.flags_toggle(Flags.down)
                self.send_ligado("Mostrando caminho das tarefas baixadas ", Flags.down, "D")
            elif value == ord("A"):
                self.flags_toggle(Flags.admin)
                self.send_ligado("Desbloqueando acesso a todas as missões ", Flags.admin, "A")
                self.reload_options()
            elif value == ord(">"):
                self.process_expand()
            elif value == ord("<"):
                self.process_collapse()
            elif value == ord(" ") or value == ord("\n"):
                self.toggle()
            elif value == ord("o"):
                self.open_link()
            elif value == ord("l"):
                self.set_language()
            elif value == ord("g"):
                self.process_down()
            elif value == ord("M"):
                self.mass_mark()
            elif value >= ord("0") and value <= ord("9"):
                self.set_grade(value)
            self.save_to_json()
        
            if self.first_loop:
                self.first_loop = False

    def play(self, graph_ext: str):
        self.graph_ext = graph_ext
        while True:
            output = curses.wrapper(self.main)
            if output is None:
                return
            else:
                output()




from typing import Optional
import shutil
from ..util.term_color import Color

class Report:
    __term_width: Optional[int] = None

    def __init__(self):
        pass

    @staticmethod
    def update_terminal_size():
        term_width = shutil.get_terminal_size().columns
        if term_width % 2 == 0:
            term_width -= 1
        Report.__term_width = term_width

    @staticmethod
    def get_terminal_size():
        if Report.__term_width is None:
            Report.update_terminal_size()

        return Report.__term_width

    @staticmethod
    def set_terminal_size(value: int):
        if value % 2 == 0:
            value -= 1
        Report.__term_width = value

    @staticmethod
    def centralize(
        text,
        sep=" ",
        left_border: Optional[str] = None,
        right_border: Optional[str] = None,
    ) -> str:
        if left_border is None:
            left_border = sep
        if right_border is None:
            right_border = sep
        term_width = Report.get_terminal_size()

        size = Color.len(text)
        pad = sep if size % 2 == 0 else ""
        tw = term_width - 2
        filler = sep * int(tw / 2 - size / 2)
        return left_border + pad + filler + text + filler + right_border
